#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill info response"""

from typing import Text, Tuple

from skill_sdk.utils.util import CamelModel
from skill_sdk.__version__ import __spi_version__


class SkillInfoResponse(CamelModel):
    """Skill information response DTO. Expected at `GET <skill-URL>/info` endpoint

    Example:
    {
      "version": "v1",
      "skillSpiVersion": "0.7",
      "skillVersion": "skill-version",
      "id": "id",
      "skillId": "skillId",
      "supportedLocales": [
        "de",
        "en",
        "fr"
      ]
    }

    """

    # Skill ID
    skill_id: Text

    # Skill version
    skill_version: Text

    # Supported locales
    supported_locales: Tuple[Text, ...] = ("de",)

    # Supported SPI version
    skill_spi_version: Text = __spi_version__



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""CLI: "translate" command"""

import argparse


def execute(args: argparse.Namespace) -> None:
    """
    Extract translations from Python modules,
    optionally can download translations from the text service

    :param args:
    :return:
    """
    from skill_sdk.config import settings
    from skill_sdk.tools.translate import download_translations, extract_translations

    #
    # If download URL specified, download translations from text services and save locally:
    # this is a step to migrate skills and decommission the service
    #
    # Once skills are migrated, this function becomes obsolete
    #
    if args.download_url:
        download_translations(
            args.download_url,
            settings.SKILL_NAME,
            args.token,
            args.tenant,
            args.force,
        )

    else:
        extract_translations(args.modules)


def add_subparser(subparsers) -> None:
    """
    Command arguments parser

    :param subparsers:
    :return:
    """

    translate_parser = subparsers.add_parser(
        "translate",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Translate a project.",
        description="Extracts translatable strings from Python modules. "
        "Creates translation catalog 'messages.pot'."
        "Optionally, downloads full catalog from text service and "
        "saves translations in YAML format as '{locale}.yaml'",
    )

    translate_parser.add_argument(
        "modules",
        nargs="*",
        default=["impl"],
        help="Modules to scan and translate (multiple values are supported, i.e.: 'impl app.py').",
    )

    translate_parser.add_argument(
        "-d",
        "--download-url",
        help="URL to download the translations catalog (text services URL).",
        action="store",
        type=str,
        nargs="?",
    )

    translate_parser.add_argument(
        "-k",
        "--token",
        help="Bearer authentication token (for the text services).",
        action="store",
        type=str,
        nargs="?",
    )

    translate_parser.add_argument(
        "-n",
        "--tenant",
        help="Tenant (for admin route authentication)",
        action="store",
        type=str,
        nargs="?",
    )

    translate_parser.add_argument(
        "-f",
        "--force",
        help="Overwrite existing translations.",
        action="store_const",
        dest="force",
        const=True,
    )

    translate_parser.set_defaults(command=execute)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Intent requests"""

from skill_sdk.intents.entities import AttributeV2

from skill_sdk.intents.request import (
    Context,
    Session,
    InvokeSkillRequest as Request,
    RequestContextVar,
    request,
    r,
)

from skill_sdk.intents.handlers import (
    intent_handler,
    invoke,
    EntityValueException,
    ErrorHandlerType,
)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#

"""Persistence service"""

import json
import logging
from typing import Any, Dict, Text

from skill_sdk.requests import HTTPError, Response
from skill_sdk.services.base import BaseService

PARSE_EXCEPTION = "%s responded with error. Data not available: %s"
REQUEST_EXCEPTION = "%s did not respond: %s"

logger = logging.getLogger(__name__)


class PersistenceService(BaseService):
    """Persistence service: a simple key-value store"""

    VERSION = 1
    NAME = "persistence"

    async def _get(self, path: str) -> Dict[Text, Any]:
        """
        Read the data from service

        :param path:
        :return:
        """

        try:
            async with self.async_client as client:
                _url = "/".join((self.url, path.strip("/\\")))

                data = await client.get(_url)
                return data.json()
        except (
            HTTPError,
            json.decoder.JSONDecodeError,
        ) as ex:
            logger.error(
                "%s responded with %s. Data not available.",
                repr(self.url),
                repr(ex),
            )
            raise

    async def get(self) -> Dict[Text, Any]:
        """
        Read the skill data

        :return:
        """
        return await self._get("entry/data")

    async def get_all(self) -> Dict[Text, Any]:
        """
        Read all the data

        :return:
        """
        return await self._get("entry")

    async def set(self, data: Dict[str, Any]) -> Response:
        """
        Update/Insert the data

        :param data:    data
        :return:
        """
        try:

            async with self.async_client as client:
                _url = f"{self.url}/entry"
                return await client.post(_url, json=dict(data=data))

        except HTTPError as ex:
            logger.error(
                "%s responded with %s. Data not changed.",
                repr(self.url),
                repr(ex),
            )
            raise

    async def delete(self) -> Response:
        """
        Delete all data

        :return:
        """
        try:

            async with self.async_client as client:
                _url = f"{self.url}/entry"
                return await client.delete(_url)

        except HTTPError as ex:
            logger.error(
                "%s responded with %s. Data not deleted.",
                repr(self.url),
                repr(ex),
            )
            raise



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Route definitions"""

import logging
import secrets
from typing import Text

from fastapi import Depends, FastAPI, Request, Security
from fastapi.responses import JSONResponse, RedirectResponse
from fastapi.exceptions import HTTPException
from fastapi.security.http import (
    HTTPBasic,
    HTTPBasicCredentials,
    HTTP_401_UNAUTHORIZED,
)

import skill_sdk.i18n
from skill_sdk.config import settings
from skill_sdk.__version__ import __version__
from skill_sdk.intents import invoke
from skill_sdk.middleware.skillassessment import SkillFitAssessmentSingleton, SkillFitnessInitiateJudgement

from skill_sdk.responses import SkillInfoResponse, SkillInvokeResponse

logger = logging.getLogger(__name__)
security = HTTPBasic()


def check_credentials(username: Text, password: Text):
    """
    Check request credentials:

        currently basic credential validation is used
        skill is invoked with user name "cvi" and a password specified as api_key in skill configuration

    :param username:
    :param password:
    :return:
    """

    def wrapper(credentials: HTTPBasicCredentials = Security(security)):
        """Check credentials and throw HTTP_401_UNAUTHORIZED if incorrect"""
        if not (
            secrets.compare_digest(credentials.username, username)
            and secrets.compare_digest(credentials.password, password)
        ):
            logger.warning("401 raised, returning: access denied.")
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail="Incorrect user name or password",
            )

    return wrapper


async def handle_info_request(request: Request):
    """
    Get skill info endpoint

        Returns the skill info to CVI:

            - skill id
            - supported SPI version
            - skill and SDK versions
            - supported locales

    :param request: Request
    :return:        JSONResponse
    """

    logger.debug("Handling info request.")

    return JSONResponse(
        SkillInfoResponse(
            skill_id=settings.SKILL_NAME,
            skill_version=f"{settings.SKILL_VERSION} {__version__}",
            supported_locales=tuple(request.app.translations.keys()),
        ).dict()
    )


async def invoke_intent(
    rq: Request,
    request: skill_sdk.intents.Request,
):
    """
    Invoke intent endpoint

        Sets the translation to requested locale and invokes an intent handler.

    :param rq:          original starlette's request
    :param request:     skill invoke request

    :return:
    """

    def _get_translation(
        app: skill_sdk.Skill, locale: Text
    ) -> skill_sdk.i18n.Translations:
        """Get translation for locale, or empty translation if does not exist"""

        if locale not in app.translations:
            logger.error("Translation for locale %s is not available.", repr(locale))
            return skill_sdk.i18n.Translations()
        return app.translations[locale]

    try:
        handler = rq.app.get_handler(request.context.intent)
    except KeyError:
        logger.error("Intent not found: %s", repr(request.context.intent))
        return JSONResponse({"code": 1, "text": "Intent not found!"}, status_code=404)

    response = await invoke(
        handler,
        request.with_translation(_get_translation(rq.app, request.context.locale)),
    )
    return JSONResponse(response.dict())


def api_base():
    """
    API base is either set directly in `skill.conf` (required for deployment as Azure function), like

        skill:
            api_base: /api

        OR constructed from skill version and name: `/v{version:1}/{name:my-skill}` => `/v1/my-skill`

        skill:
            version: 1
            name: my-skill

    """

    return (
        settings.API_BASE
        if settings.API_BASE is not None
        else f"/v{settings.SKILL_VERSION}/{settings.SKILL_NAME}"
    )


async def skill_fit_assessment_init_setup(r: SkillFitnessInitiateJudgement) -> JSONResponse:
    skill_fit_assessment = SkillFitAssessmentSingleton.instance()
    return skill_fit_assessment.get_response_implementation(r)


async def health(r: Request) -> JSONResponse:
    """
    Health check endpoint

    :param r:   starlette's request

    :return:
    """
    if len(r.app.intents) < 1:
        return JSONResponse(dict(text="No intent handlers loaded!"), status_code=418)

    return JSONResponse(dict(text="Ok"), status_code=200)


def setup_routes(app: FastAPI):
    """
    Setup default skill routes:

        - GET   /info
        - POST  /
        - GET   /k8s/readiness
        - GET   /k8s/liveness

    :param app:
    :return:
    """

    authentication = (
        [Depends(check_credentials(settings.SKILL_API_USER, settings.SKILL_API_KEY))]
        if not app.debug
        else []
    )

    app.add_api_route(
        f"{api_base()}/info",
        handle_info_request,
        dependencies=authentication,
        response_model=SkillInfoResponse,
        name="Get Skill Info",
        tags=["Skill endpoints"],
    )

    app.add_api_route(
        f"{api_base()}",
        invoke_intent,
        dependencies=authentication,
        methods=["POST"],
        response_model=SkillInvokeResponse,
        response_model_exclude_none=True,
        name="Invoke Intent",
        tags=["Skill endpoints"],
    )

    app.add_api_route(
        f"{api_base()}" + settings.SKILL_FIT_ASSESSMENT_ENDPOINT,
        skill_fit_assessment_init_setup,
        dependencies=authentication,
        methods=["POST"],
        name="Skill Fit Assessment",
        tags=["Skill endpoints"],
    )

    app.add_route(
        settings.K8S_READINESS,
        health,
        name="Readiness Probe",
    )

    app.add_route(
        settings.K8S_LIVENESS,
        health,
        name="Liveness Probe",
    )

    # Redirect root to "/redoc", if not in "debug" mode
    if not app.debug:
        app.add_route("/", RedirectResponse(url=app.redoc_url or "/redoc"))



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Middleware definitions"""

from starlette_context.middleware import RawContextMiddleware
from starlette_context import context  # noqa

from skill_sdk.middleware import (
    error,
    log,
)

from skill_sdk.middleware.log import HeaderKeys


def setup_middleware(app):
    """
    Setup default middlewares: error and context

    :param app:
    :return:
    """

    error.setup_middleware(app)

    app.add_middleware(
        RawContextMiddleware,
        plugins=(
            log.TraceIdPlugin(),
            log.SpanIdIdPlugin(),
            log.TenantIdIdPlugin(),
            log.TestingFlagPlugin(),
            log.UserDebugLogPlugin(),
            log.MagentaTransactionId(),
        ),
    )

    # Since Prometheus metrics exporter is optional,
    # try to load prometheus middleware and simply eat an exception
    try:
        from skill_sdk.middleware.prometheus import setup

        setup(app)
    except ModuleNotFoundError:
        pass



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""CLI: "version" command"""

#
# Simply outputs the skill version as specified in config files:
#
#   this command is used in internal GitLab pipelines
#   to generate docker image name for the skill
#

import argparse
import logging

from skill_sdk.cli import (
    add_env_file_argument,
    add_module_argument,
    process_env_file,
)


def execute(arguments):
    """Print skill version"""

    from skill_sdk import config, log

    process_env_file(arguments)

    # Set default log level to ERROR, if not explicitly overridden with "--verbose"/"--debug"
    loglevel = getattr(arguments, "loglevel", None) or logging.ERROR
    log.setup_logging(loglevel, config.FormatType.HUMAN)

    # Reload setting from a config file
    config.settings.reload()

    print(config.settings.SKILL_VERSION)


def add_subparser(subparsers):
    """
    Command arguments parser

    :param subparsers:
    :return:
    """

    version_parser = subparsers.add_parser(
        "version",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Print skill version and exit.",
    )
    add_env_file_argument(version_parser)
    add_module_argument(version_parser)
    version_parser.set_defaults(command=execute)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Utility functions"""

import re
import sys
import time
import asyncio
import inspect
import datetime
import threading
import contextlib
import unittest.mock
import importlib.util
from functools import partial
from contextvars import copy_context
from concurrent.futures import ThreadPoolExecutor
from types import ModuleType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Text,
    TypeVar,
    Union,
)

import orjson
import pydantic
import nest_asyncio
from pydantic import root_validator, validator, ValidationError  # noqa
from pydantic.utils import lenient_issubclass
import uvicorn


T = TypeVar("T")
DEFAULT_LOCALE = "de"


def camel_to_snake(name):
    """
    Convert CamelCase to snake_case

    :param name:
    :return:
    """
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def snake_to_camel(name):
    """
    Convert snake_case to CamelCase

    :param name:
    :return:
    """
    reg_ex = re.compile(r"_+[a-z0-9]")
    return reg_ex.sub(lambda x: x.group(0)[1].upper(), name)


def orjson_dumps(v, *, default):
    """
    orjson.dumps returns bytes, to match standard json.dumps we need to decode

    :param v:
    :param default:
    :return:
    """
    return orjson.dumps(v, default=default).decode()


class Server(uvicorn.Server):
    """Uvicorn server implementing "run_in_thread" context manager:

    >>> server = Server("app:app", port=4242)
    >>> with server.run_in_thread():
    >>>     # Server started.
    >>>     ...
    >>> # Server stopped

    """

    WAIT_TIME = 1e-3

    def __init__(self, *args, **kwargs):
        if "loop" not in kwargs:
            kwargs.update(dict(loop="asyncio"))
        config = uvicorn.Config(*args, **kwargs)
        super().__init__(config)

    @contextlib.contextmanager
    def run_in_thread(self):
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(Server.WAIT_TIME)
            yield
        finally:
            self.should_exit = True
            thread.join()


def reload_recursive(module: ModuleType) -> ModuleType:
    """
    Recursively reload a runner module (app.py) with submodules

    :param module:
    :return:
    """
    reloaded = set()

    def reload_parent(module_name):
        parent = sys.modules[module_name]
        if parent not in reloaded:
            reloaded.add(parent)
            importlib.reload(parent)

    def visit(m):
        reloaded.add(m)
        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if isinstance(attr, Callable):
                reload_parent(attr.__module__)
            if isinstance(attr, ModuleType):
                name = attr.__name__
                dot = name.rfind(".")
                if dot > 0:
                    reload_parent(name[:dot])
                elif attr not in reloaded:
                    visit(attr)
        importlib.reload(m)
        return m

    return visit(module)


class ContextVarExecutor(ThreadPoolExecutor):
    """Copy existing contextVars before executing"""

    def submit(self, *args, **kwargs):
        ctx = copy_context()

        return super().submit(ctx.run, *args, **kwargs)


async def run_in_executor(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Run a synchronous function in a thread pool to prevent blocking

    :param func:
    :param args:
    :param kwargs:
    :return:
    """
    loop = asyncio.get_running_loop()

    return await loop.run_in_executor(
        ContextVarExecutor(), partial(func, *args, **kwargs)
    )


def run_until_complete(func: Awaitable[T]) -> T:
    """
    Run an asynchronous function in synchronous context

    :param func:
    :return:
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    if loop.is_running():
        nest_asyncio.apply(loop)

    return loop.run_until_complete(func)


class BaseModel(pydantic.BaseModel):
    """
    Override Pydantic's BaseModel defaults:

        - use faster orjson for objects serialization

        - make all instances immutable
            (with an exception of `responses.Result` - that is mutable for backward compatibility)

        - allow objects population by the model attribute

        - use enum value instead of raw enums when populating properties

    """

    class Config:
        """Alter default config for base models"""

        # Use orjson to serialize
        json_dumps = orjson_dumps

        # Make all model instances "faux-immutable" by default
        allow_mutation = False

        # Allow objects population by the model attribute, as well as the alias
        allow_population_by_field_name = True

        # Populate models with the value property of enums, rather than the raw enum
        use_enum_values = True

    def dict(self, *args, **kwargs) -> Dict[Text, Any]:
        """
        Alter parent defaults when exporting:

            - exclude "None" values
            - export fields by alias

        """
        params = {
            "exclude_none": kwargs["exclude_none"]
            if "exclude_none" in kwargs
            else True,
            "by_alias": kwargs["by_alias"] if "by_alias" in kwargs else True,
        }

        return super().dict(*args, **{**kwargs, **params})


class CamelModel(BaseModel):
    """CamelModel will use camelCase aliases for snake_case fields"""

    class Config:
        """Set `snake_to_camel` as alias generator function"""

        alias_generator = snake_to_camel


def attrs_examples(intent: Callable) -> Dict[str, List]:
    """
    Create attribute samples for intent

    :param intent:
    :return:
    """

    from skill_sdk.intents import AttributeV2, Request

    # Extract parameter arguments from intent implementation
    parameters = {
        name: param
        for name, param in inspect.signature(intent).parameters.items()
        if not lenient_issubclass(param.annotation, Request)
    }

    # 'timezone' is always present
    if "timezone" not in parameters:
        parameters["timezone"] = inspect.Parameter(
            "timezone", inspect.Parameter.KEYWORD_ONLY, default="UTC"
        )

    def value(param):
        """Try to set value to an attribute"""
        return (
            param.default
            if param.default and param.default != inspect.Parameter.empty
            else "value"
        )

    attrs_v2 = {
        name: [AttributeV2(value=value(param), id=0).dict()]
        for name, param in parameters.items()
    }

    return attrs_v2


#
#   Helpers for Swagger UI/Redoc
#


def intent_examples(intents: Mapping[Text, Callable]) -> Dict[str, Dict]:
    """
    Create example intent calls

    :param intents: List of intents
    :return:
    """
    examples = {
        f"{snake_to_camel(name)}Example": {
            "summary": name,
            "value": {
                "context": {
                    "intent": name,
                    "attributesV2": attrs_examples(intent),
                    "configuration": {},
                    "locale": DEFAULT_LOCALE,
                    "tokens": {},
                },
                "session": {
                    "id": 123,
                    "new": True,
                    "attributes": {"attr-1": 1, "attr-2": "2"},
                },
            },
        }
        for name, intent in intents.items()
    }

    return examples


def populate_intent_examples(intents: Mapping[Text, Callable]):
    """
    Create intent invoke examples for Swagger UI

    :param intents:
    :return:
    """
    from skill_sdk.intents import Context

    Context.__config__.schema_extra = {"examples": intent_examples(intents)}


#
#   Unit test helpers:
#

datetime_class = datetime.datetime


def mock_datetime_now(target, dt):
    """
    Patch datetime.datetime.now with __instancecheck__
    """

    class DatetimeSubclassMeta(type):
        """Metaclass to mock __instancecheck__"""

        @classmethod
        def __instancecheck__(mcs, obj):  # NOSONAR
            return isinstance(obj, datetime_class)

    class BaseMockedDatetime(datetime_class):
        """Patch the original datetime.datetime.now"""

        @classmethod
        def now(cls, tz=None):
            return target.replace(tzinfo=tz)

        @classmethod
        def utcnow(cls):
            return target

    dt_mock = DatetimeSubclassMeta("datetime", (BaseMockedDatetime,), {})
    return unittest.mock.patch.object(dt, "datetime", dt_mock)


def mock_date_today(now):
    """
    Patch datetime.date.today
    """

    class MockedDate(datetime.date):
        @classmethod
        def today(cls):
            return now

    return unittest.mock.patch.object(datetime, "date", MockedDate)


def create_request(
    intent: Text, session: Union[CamelModel, Dict[Text, Text]] = None, **kwargs
):
    """
    Skill invoke request factory: used in unit tests to mock a skill invoke request

    :param intent:      Intent name
    :param session:     Session (or session attributes)
    :param kwargs:      Keyword arguments are forwarded to create_context factory
    :return:
    """

    from skill_sdk.__version__ import __spi_version__
    from skill_sdk.intents import Request, Session
    from skill_sdk.responses import SessionResponse

    example: Dict = Session.Config.schema_extra["example"]
    if isinstance(session, Session):
        _session = session.copy()
    elif isinstance(session, SessionResponse):
        _session = Session(
            **{
                **dict(id=example["id"], new=False),
                **dict(attributes=session.attributes),
            }
        )
    elif isinstance(session, Dict):
        _session = Session(
            **{**dict(id=example["id"], new=example["new"]), **dict(attributes=session)}
        )
    else:
        _session = Session(**example)

    return Request(
        context=create_context(intent=intent, **kwargs),
        session=_session,
        spi_version=__spi_version__,
    )


def test_request(
    intent: Text,
    session: Union[CamelModel, Dict[Text, Text]] = None,
    translation=None,  # noqa: don't want to import i18n for type hinting `Translations`
    **kwargs,
):
    """
    Context manager to setup global request object for unit tests

    :param intent:      Intent name for the test request
    :param session:     Session attributes
    :param translation: Optional translations
    :param kwargs:      The rest of kwargs goes to `create_context`

    :return:
    """
    from skill_sdk.i18n import Translations
    from skill_sdk.intents.request import RequestContextVar

    translations = kwargs.pop("translation", Translations())

    request = create_request(intent, session=session, **kwargs).with_translation(
        translations
    )

    return RequestContextVar(request=request)


def create_context(
    intent: Text,
    *,
    configuration: Dict[Text, Dict] = None,
    locale: Text = DEFAULT_LOCALE,
    skill_id: Text = None,
    tokens: Dict[Text, Text] = None,
    client_type_name: Text = None,
    user_profile_config: Text = None,
    **kwargs,
):
    """
    Intent invoke context factory:
    Creates skill invocation context with attributes supplied as keyword arguments.

    :param intent:      Intent name
    :param configuration:       Skill configuration
    :param locale:      Request locale
    :param skill_id:    Optional skill ID
    :param tokens:      Intent tokens (if requires)
    :param client_type_name:    Client type (mini/premium/etc)
    :param user_profile_config: User profile configuration
    :param kwargs:      Attributes(V2)
    :return:
    """

    from skill_sdk.intents import Context, AttributeV2

    def parse(_: Any) -> AttributeV2:
        """Try to create AttributeV2 from any string"""

        try:
            attr_v2: AttributeV2 = AttributeV2(_)
        except (AttributeError, TypeError, ValidationError, ValueError):
            attr_v2 = AttributeV2({"id": 0, "value": _})
        return attr_v2

    attributes_v2 = {
        key: [parse(each) for each in value]
        if isinstance(value, (list, tuple))
        else [parse(value)]
        for key, value in kwargs.items()
        if value is not None
    }

    # 'timezone' is always present
    if "timezone" not in attributes_v2:
        attributes_v2["timezone"] = [parse("Europe/Berlin")]

    # Create `attributes` dict for backward compatibility
    attributes: Dict[Text, List] = {
        key: [item.value for item in value] for key, value in attributes_v2.items()
    }

    return Context(
        attributes=attributes or {},
        attributes_v2=attributes_v2 or {},
        configuration=configuration or {},
        intent=intent,
        locale=locale,
        skill_id=skill_id,
        tokens=tokens or {},
        client_type_name=client_type_name,
        user_profile_config=user_profile_config,
    )



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Command line interface"""

import argparse

from skill_sdk.cli import (
    add_logging_options,
    init,
    develop,
    run,
    translate,
    version,
)


def main() -> None:
    """
    CLI - run as a standalone python app

    :return:
    """

    parser = argparse.ArgumentParser(
        prog="vs",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Magenta Voice Skill CLI. 'vs' command is a helper that "
        "allows you to execute Magenta skill tasks: "
        "initialize skill project, define or test your intent handlers.",
    )

    add_logging_options(parser)
    subparsers = parser.add_subparsers(dest="command", help="Skill commands")

    # Initialize a skill project
    init.add_subparser(subparsers)

    # Run the skill with uvicorn server
    run.add_subparser(subparsers)

    # Run the skill in development mode (with Designer UI)
    develop.add_subparser(subparsers)

    # Extracts translatable strings from Python modules
    translate.add_subparser(subparsers)

    # Output skill version
    version.add_subparser(subparsers)

    arguments = parser.parse_args()

    if getattr(arguments, "command", None):
        arguments.command(arguments)
    else:
        parser.exit(-1, parser.format_help())


if __name__ == "__main__":
    main()



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#

"""**DEPRECATED**: Text service"""

import json
import logging
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Text
from datetime import datetime

from skill_sdk import i18n
from skill_sdk.utils.util import CamelModel, validator
from skill_sdk.requests import HTTPError
from skill_sdk.services.base import BaseService

logger = logging.getLogger(__name__)


class Locale(CamelModel):
    """Language code in ISO 639-1 format"""

    code: Text


class LocaleInfo(CamelModel):
    """List of supported languages"""

    supported_languages: List[Locale]

    @validator("supported_languages", each_item=True)
    def check_lang_regex(cls, v: Locale):  # pylint: disable=E0213
        if i18n.RE_TRANSLATIONS.match(v.code):
            return v


class Translation(CamelModel):
    """Text service translation record"""

    # Language
    locale: Text

    # Scope (usually skill name)
    scope: Text

    # Translation tag
    tag: Text

    # Translations (possibly, multiple)
    sentences: List[Text]

    # Tenant, the record is valid for
    tenant: Text

    comment: Optional[Text]
    last_change_date: Optional[datetime]
    creation_date: Optional[datetime]


TranslationCatalog = Dict[Text, Dict[Text, List[Text]]]


class TextService(BaseService):
    """Text (translation) service: stripped version to download full translation catalogue"""

    VERSION = 1
    NAME = "text"

    def __init__(self, url: Text, scope: Text, **kwargs) -> None:
        self.url = url
        self.scope = scope
        super().__init__(url, **kwargs)

    def admin_get_full_catalog(self) -> TranslationCatalog:
        """
        Get a complete translations catalog as {"language": {"key": "value"}} dictionary (admin route only!)

        :return:
        """
        catalog: DefaultDict = defaultdict(dict)

        try:

            logger.debug("Getting locale info...")
            data = self.client.get(f"{self.url}/info/locale").json()
            info = LocaleInfo(**data)
            logger.debug("Locale info: %s", info.dict())

            data = self.client.get(f"{self.url}/scope/{self.scope}").json()
            entries = [Translation(**_) for _ in data]

            [catalog[_.locale].update({_.tag: _.sentences}) for _ in entries]

        except (
            KeyError,
            TypeError,
            HTTPError,
            json.decoder.JSONDecodeError,
        ) as ex:
            logger.error(
                "%s responded with %s. Catalog not available.",
                repr(self.url),
                repr(ex),
            )

        return dict(catalog)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""CLI: "run" command"""

import argparse
import logging
from contextlib import closing

import uvicorn

from skill_sdk.cli import (
    add_env_file_argument,
    add_module_argument,
    import_module_app,
    process_env_file,
)

logger = logging.getLogger(__name__)


def execute(arguments):
    """
    Initialize and run an app with "uvicorn" server

    :return:
    """
    from skill_sdk import config, log

    process_env_file(arguments)

    # Set default log level to WARNING, if not explicitly overridden with "--verbose"/"--debug"
    loglevel = getattr(arguments, "loglevel", None) or logging.WARNING
    log.setup_logging(loglevel)

    module, app = import_module_app(arguments.module)

    logger.info("Loaded app: %s", repr(app))

    if not app.intents:
        raise RuntimeError(
            "No intent handlers loaded. Check the log messages for import errors..."
        )

    logger.info("Loaded handlers: %s", list(app.intents))

    run_config = config.settings.http_config()

    logger.info("Starting app with config: %s", repr(run_config))

    with closing(app):
        uvicorn.run(app, **run_config)


def add_subparser(subparsers):
    """
    Command arguments parser

    :param subparsers:
    :return:
    """

    run_parser = subparsers.add_parser(
        "run",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Starts the skill in production mode.",
        description="Run the HTTP server as configured to handle requests.",
    )
    add_env_file_argument(run_parser)
    add_module_argument(run_parser)
    run_parser.set_defaults(command=execute)



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(HERE, "skill_sdk", "__version__.py")) as f:
    exec(f.read(), about)


setup(
    name=about["__name__"],
    version=about["__version__"],
    description=about["__description__"],
    url=about["__url__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    license=about["__license__"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    packages=find_packages(),
    package_data={
        "": ["*"],
        "skill_sdk.cli": [
            "scaffold/*",
            "scaffold/impl/*",
            "scaffold/locale/*",
            "scaffold/scripts/*",
            "scaffold/tests/*",
        ],
        "skill_sdk.ui": ["css/*", "js/*", "templates/*"],
    },
    install_requires=[
        "fastapi>=0.70.0,<1",
        "pydantic>=1.8,<2.0.0",
        "starlette_context",
        "python-dateutil",
        "babel",
        "uvicorn[standard]",
        "isodate",
        "orjson",
        "aiobreaker",
        "httpx>=0.16, <1",
        "pyyaml",
        "nest-asyncio",
        "pycryptodome",
        "Crypto~=1.4.1"
    ],
    extras_require={
        "dev": [
            "gunicorn",
            "starlette[full]",
            "mypy>=0.9",
            "respx",
            "black",
            "pytest",
            "pytest-cov",
            "pytest-mock",
            "pytest-asyncio",
            "questionary",
            "starlette-opentracing",
            "jaeger-client",
            "starlette-exporter",
            "types-orjson",
            "types-python-dateutil",
            "types-pkg_resources",
            "types-PyYAML"
        ],
        "all": [
            "starlette-opentracing",
            "starlette-exporter",
        ],
    },
    entry_points={"console_scripts": ["vs = skill_sdk.__main__:main"]},
    setup_requires=["wheel"],
    python_requires=">=3.7",
)



#
# voice-skill-sdk
#
# (C) 2022, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#


import base64
import json
import logging
from typing import Any, Dict

from skill_sdk.intents import r
from skill_sdk.config import settings
from Crypto.Cipher import AES

logger = logging.getLogger(__name__)


class ServiceTokenDecryption:
    """
    Utility class responsible for decrypting the cv service-token.

    NOTE :: Please do not store the user related token claims
            in any permanent persistence storage due to GDR compliance issues.
            Currently there is not any existing mechanism for cleaning up any user related data.
    """

    def __init__(self, service_token: str = "", service_token_secret: str = "") -> None:
        self.service_token = service_token
        self.service_token_secret = service_token_secret

    def decrypt(self) -> Dict:
        """
        Verifies and decrypts the cvi service-token.
        Returns the claims of the token

        :param service_token: 
            The token that should be decrypted

        :param service_token_secret: 
            The secret to decrypt the service token

        :return:
            the claims of the token

        :Raises ValueError:
            if the MAC does not match. The message has been tampered with
            or the key is incorrect.
        """

        if len(self.service_token) == 0:
            e = ValueError("Service Token is empty")
            logger.error(e)
            raise e

        if len(self.service_token_secret) == 0:
            e = ValueError("Service Token Secret is empty")
            logger.error(e)
            raise e

        decoded_cvi_token: bytes = base64.b64decode(self.service_token)
        decoded_str = str(decoded_cvi_token, "utf-8")
        decoded_dict: dict = json.loads(decoded_str)
        decoded_nonce = base64.b64decode(str(decoded_dict.get("nonce")))
        decoded_encrypted_plain_token = base64.b64decode(
            str(decoded_dict.get("encryptedPlainToken"))
        )
        decoded_secret: bytes = base64.b64decode(self.service_token_secret)

        cipher: Any = AES.new(key=decoded_secret, mode=AES.MODE_GCM, nonce=decoded_nonce)   # Setup cipher
        # https://docs.python.org/3/library/typing.html#the-any-type
        # Prevent mypy to miss judge the return value of the AES.new method by making sure the object will be marked as Any

        # Raw data structure for encryptedPlainToken crypttext[:-16] + auth_tag[-16:] aka default cipher.block_size
        decoded_cipher_text = decoded_encrypted_plain_token[: -cipher.block_size]
        decoded_auth_tag = decoded_encrypted_plain_token[-cipher.block_size :]
        try:
            plaintext = cipher.decrypt_and_verify(
                decoded_cipher_text,
                decoded_auth_tag,
            )
            logger.debug(str(plaintext, "utf-8"))
        except ValueError as e:
            logger.error(e)
            raise e
        token_data = json.loads(str(plaintext, "utf-8"))
        
        return token_data



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill invoke request"""

import datetime
import logging
from contextlib import ContextDecorator
from contextvars import ContextVar, Token
from typing import Any, Dict, List, Optional, Text
from dateutil import tz

from pydantic import Field

from skill_sdk.__version__ import __spi_version__
from skill_sdk.utils.util import CamelModel, DEFAULT_LOCALE
from skill_sdk.intents import AttributeV2
from skill_sdk.i18n import _, _n, _a, Translations

logger = logging.getLogger(__name__)


class Context(CamelModel):
    """Intent invocation context"""

    # Skill attributes
    attributes: Dict[Text, List[Text]]

    # Skill attributes in extended form
    attributes_v2: Dict[Text, List[AttributeV2]]

    # Skill configuration
    configuration: Dict[Text, Dict]

    # Intent name
    intent: Text

    # Request locale
    locale: Text = Field(example=DEFAULT_LOCALE)

    # Skill id
    skill_id: Optional[Text]

    # Skill tokens
    tokens: Dict[Text, Text]

    # Client type (mini/premium/etc)
    client_type_name: Optional[Text]

    # User profile configuration
    user_profile_config: Optional[Text]

    @staticmethod
    def _(*args, **kwargs):
        return _(*args, **kwargs)

    @staticmethod
    def _n(*args, **kwargs):
        return _n(*args, **kwargs)

    @staticmethod
    def _a(*args, **kwargs):
        return _a(*args, **kwargs)

    def _get_attr_value(self, attr, default=None):
        """Silently return first item from attributes array"""
        try:
            attr_v2 = self.attributes_v2[attr][0]
            return attr_v2.value
        except LookupError:
            logger.error(
                "Attribute %s is not in context. Defaulting to %s",
                repr(attr),
                repr(default),
            )
            return default

    def gettz(self) -> datetime.tzinfo:
        """
        Get device timezone from context attributes

        :return:
        """
        _tz = self._get_attr_value("timezone")
        timezone = tz.gettz(_tz)

        if timezone is None:
            logger.error(
                "Device timezone not present or invalid: %s. Defaulting to UTC",
                repr(_tz),
            )
            timezone = tz.tzutc()

        return timezone

    def today(self) -> datetime.datetime:
        """
        Get `datetime.datetime` object representing the current day at midnight

        :return:
        """
        dt = self.now()
        return datetime.datetime.combine(dt.date(), datetime.time(0))

    def now(self) -> datetime.datetime:
        """Get current device date/time with timezone info

        :return:
        """
        timezone = self.gettz()
        return datetime.datetime.now(datetime.timezone.utc).astimezone(timezone)


class Session(CamelModel):
    """Device session"""

    class Config:
        """Sample values for Swagger UI"""

        schema_extra: Dict = {
            "example": {
                "id": "123",
                "attributes": {"key-1": "value-1", "key-2": "value-2"},
                "new": True,
            }
        }

        # Session is mutable for backward compatibility
        allow_mutation = True

    # Session identifier
    id: Text

    # Session attributes
    attributes: Dict[Text, Text]

    # True if session is new, False for resumed
    new: bool = True

    def __getitem__(self, item):
        return self.attributes.__getitem__(item)

    def __setitem__(self, key, value):
        return self.attributes.__setitem__(key, value)

    def __delitem__(self, key):
        return self.attributes.__delitem__(key)


class InvokeSkillRequest(CamelModel):
    """Skill invocation request"""

    # Invocation context
    context: Context

    # Session
    session: Session

    # Service Provider Interface (SPI) version
    spi_version: Text = Field(example=__spi_version__)

    # Translations to current locale
    _trans: Optional[Translations]

    def get_translation(self) -> Optional[Translations]:
        return self._trans

    def with_translation(self, translation: Translations) -> "InvokeSkillRequest":
        """
        Factory method to add translation to current locale

        :param translation:
        :return:
        """
        return self.copy(update=dict(_trans=translation))


class RequestContextVar(ContextDecorator):
    """Context manager to make InvokeSkillRequest object globally importable"""

    _scope = "request"
    _request_scope_storage: ContextVar[Dict[Text, Any]] = ContextVar(
        "invoke_skill_request"
    )

    __request_token: Optional[Token] = None

    def __init__(self, **kwargs: Any):
        # Deep-copy request to allow mutating session attributes
        self.kwargs = {k: v.copy(deep=True) for k, v in kwargs.items()}

    def __len__(self):
        return 0 if self.__request_token is None else 1

    def __getattr__(self, item):
        try:
            return getattr(self._request_scope_storage.get()[self._scope], item)
        except (AttributeError, LookupError) as e:
            logger.error(
                "Accessing request local object outside of the request-response cycle."
            )
            logger.debug("Details: %s", repr(e))

    def __enter__(self):
        self.__request_token = self._request_scope_storage.set(self.kwargs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._request_scope_storage.reset(self.__request_token)
        if exc_type:
            return False


def _context_var():
    """Wrapper to cheat mypy"""
    return RequestContextVar()


#
# Global objects
#
request: InvokeSkillRequest = _context_var()  # noqa

r = request



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""CLI: "develop" command"""

#
#   - starts skill-designer
#   - starts skill in development mode
#   - displays skill logs in "Test -> log output"
#   - modifies Python sources in "Design"
#

import os
import argparse
import logging
import pathlib
from contextlib import closing

import uvicorn

from skill_sdk.cli import (
    add_env_file_argument,
    add_module_argument,
    import_module_app,
    process_env_file,
    DEFAULT_MODULE,
)

logger = logging.getLogger(__name__)


def execute(arguments):
    """
    Run the app in "development" mode:

        set logging level to DEBUG
        logging format to human-readable
        override debug flag from configuration
        initialize Designer UI

    :param arguments:
    :return:
    """
    from skill_sdk import config, log

    # Set "debug" mode to prevent installing check_credentials dependency
    os.environ["SKILL_DEBUG"] = "true"
    process_env_file(arguments)

    # Set default log level to DEBUG, if not explicitly overridden with "--verbose"/"--quiet"
    loglevel = getattr(arguments, "loglevel", None) or logging.DEBUG
    log.setup_logging(loglevel, config.FormatType.HUMAN)

    if arguments.module == DEFAULT_MODULE:
        create_if_missing(arguments.module)

    module, app = import_module_app(arguments.module)

    logger.info("Loaded app: %s", repr(app))
    logger.info("Loaded handlers: %s", list(app.intents))

    run_config = config.settings.http_config()
    logger.info("Starting app with config: %s", repr(run_config))

    with closing(app.develop()):
        uvicorn.run(app, **run_config)


def create_if_missing(module) -> None:
    """
    When `vs develop` is called from an empty project,
    "impl" folder does not exist and "import_module_app" fails with ModuleNotFoundError.

        We'll double check if "impl"/"impl.py" is missing and create an empty folder.

    :param module:
    :return:
    """
    module_path = pathlib.Path.cwd() / module
    if module_path.exists() or module_path.with_suffix(".py").exists():
        return

    logger.debug("Creating modules folder %s...", module_path)
    module_path.mkdir()


def add_subparser(subparsers):
    """
    Command arguments parser

    :param subparsers:
    :return:
    """

    run_parser = subparsers.add_parser(
        "develop",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Starts the skill in development mode.",
    )
    add_env_file_argument(run_parser)
    add_module_argument(run_parser)
    run_parser.set_defaults(command=execute)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Re-prompt response"""

from typing import Optional, Text, Dict, Any
from pydantic import validator

from skill_sdk.intents import request

from skill_sdk.responses.response import (
    SkillInvokeResponse,
    ResponseType,
)


class Reprompt(SkillInvokeResponse):
    """
    Re-prompt response is sent to user as a measure to limit a number of re-prompts

    """

    # Response type defaults to ASK
    type: ResponseType = ResponseType.ASK

    # stop text will be sent if number of re-prompts is higher than maximum number
    stop_text: Optional[Text]

    # entity name if re-prompt is used for intent/entity
    entity: Optional[Text]

    # maximum number of re-prompts
    max_reprompts: int

    @validator("max_reprompts")
    def check_max_reprompts(cls, max_reprompts, values):
        # Name of the counter formatted as INTENT_ENTITY_reprompt_count
        entity = values.get("entity")
        name = (
            f"{request.context.intent}{'_' + entity if entity else ''}_reprompt_count"
        )

        try:
            reprompt_count = int(request.session.attributes[name]) + 1
        except (KeyError, ValueError):
            reprompt_count = 1

        if reprompt_count > max_reprompts > 0:
            del request.session.attributes[name]
            values["text"] = values["stop_text"]
            values["type"] = ResponseType.TELL
        else:
            request.session.attributes[name] = reprompt_count

        return max_reprompts

    def __init__(
        self,
        text: Text,
        stop_text: Text = None,
        max_reprompts: int = 0,
        entity: Text = None,
        **kwargs,
    ) -> None:
        super().__init__(
            text=text,
            stop_text=stop_text,
            max_reprompts=max_reprompts,
            entity=entity,
            **kwargs,
        )

    def dict(self, *args, **kwargs) -> Dict[Text, Any]:
        return super(Reprompt, self).dict()
